// Copyright (C) 1999-2007
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;

#include "gzip.h"
#include "zlib.h"

#define ASCII_FLAG   0x01 /* bit 0 set: file probably ascii text */
#define HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define COMMENT      0x10 /* bit 4 set: file comment present */
#define RESERVED     0xE0 /* bits 5..7: reserved */

template<class T> FitsGzipm<T>::FitsGzipm(FitsFile* fits) 
  : FitsCompressm<T>(fits)
{
  // create header
  if (!FitsCompressm<T>::initHeader(fits))
    return;

  // uncompress
  if (!inflate(fits))
    return;

  // swap if needed
  FitsCompressm<T>::swapBytes();

  // all done
  FitsCompressm<T>::valid_ = 1;
}

template <class T> int FitsGzipm<T>::inflate(FitsFile* fits)
{
  FitsHead* srcHead = fits->head();
  FitsTableHDU* srcHDU = (FitsTableHDU*)(srcHead->hdu());

  T* dest = new T[FitsCompressm<T>::size_];
  if (!dest) {
    if (DebugCompress)
      cerr << "GZIP Internal Error: unable to allocate memory" << endl;
    return 0;
  }

  // init image
  memset(dest, 0, FitsCompressm<T>::size_*sizeof(T));

  // src
  char* sdata = (char*)fits->data(); 
  char* sptr = sdata;
  int heap = srcHDU->realbytes();
  int rowlen = srcHDU->width();
  int rows = srcHDU->rows();

  // dest
  T* dptr = dest;
  int ocnt = FitsCompressm<T>::ww_*FitsCompressm<T>::hh_*FitsCompressm<T>::dd_;

  for (int ii=0; ii<rows; ii++, sptr+=rowlen) {
    // for memory models that support internal paging
    // all we need is just one row
    sptr = fits->page(sptr, rowlen);

    double zs = FitsCompressm<T>::bscale_;
    if (FitsCompressm<T>::zscale_)
      zs = FitsCompressm<T>::zscale_->value(sptr,0);

    double zz = FitsCompressm<T>::bzero_;
    if (FitsCompressm<T>::zzero_)
      zz = FitsCompressm<T>::zzero_->value(sptr,0);

    if (FitsCompressm<T>::compress_) {
      int icnt=0;
      unsigned char* ibuf = 
	(unsigned char*)((FitsBinColumnArray*)FitsCompressm<T>::compress_)->get(sdata+heap, sptr, &icnt);
      int obuf[ocnt];
      memset(obuf,0,ocnt*sizeof(int));

      z_stream zstrm;
      zstrm.next_in = NULL;
      zstrm.avail_in = 0;
      zstrm.zalloc = NULL;
      zstrm.zfree = NULL;
      zstrm.opaque = NULL;

      // look for both zlib and gzip headers
      if (inflateInit2(&zstrm, MAX_WBITS+32) != Z_OK) {
	if (DebugCompress)
	  cerr << "inflateInit error" << endl;
	return 0;
      }

      zstrm.avail_in = icnt;
      zstrm.next_in = ibuf;
      zstrm.avail_out = ocnt*sizeof(int);
      zstrm.next_out = (Bytef*)obuf;

      if (DebugCompress)
	cerr << "  inflate START: avail_in " 
	     << zstrm.avail_in
	     << " avail_out " << zstrm.avail_out
	     << " total_in " << zstrm.total_in 
	     << " total_out " << zstrm.total_out << endl;

      int result = ::inflate(&zstrm, Z_FINISH);

      switch (result) {
      case Z_OK:
      case Z_STREAM_END:
	if (DebugCompress)
	  cerr << "  inflate STREAM_END: avail_in " 
	       << zstrm.avail_in
	       << " avail_out " << zstrm.avail_out
	       << " total_in " << zstrm.total_in 
	       << " total_out " << zstrm.total_out << endl;
	break;
      case Z_BUF_ERROR:
	if (DebugCompress)
	  cerr << "  inflate BUF_ERROR: avail_in " << zstrm.avail_in
	       << " avail_out " << zstrm.avail_out << endl;
	//	return 0;
	break;
      default:
	if (DebugCompress)
	  cerr << "  inflate error " << result << endl;
	return 0;
      }

      if (zs != 1 || zz != 0)
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk]*zs + zz;
      else
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk];

      inflateEnd(&zstrm);
    }

    if (FitsCompressm<T>::uncompress_) {
      int ocnt=0;
      T* obuf = 
	(T*)(((FitsBinColumnArray*)FitsCompressm<T>::uncompress_)->get(sdata+heap, sptr, &ocnt));

      if (zs != 1 || zz != 0)
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk]*zs + zz;
      else
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk];
    }
  }

  // for memory models that support internal paging
  fits->resetpage();

  FitsCompressm<T>::data_ = dest;
  return 1;
}

template class FitsGzipm<unsigned char>;
template class FitsGzipm<short>;
template class FitsGzipm<unsigned short>;
template class FitsGzipm<int>;
template class FitsGzipm<long long>;
template class FitsGzipm<float>;
template class FitsGzipm<double>;
