// Copyright (C) 1999-2007
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;

#include "hcompress.h"
#include "util.h"

extern "C" {
  int dodecode(unsigned char**, int*, int, int, unsigned char*);
  void undigitize(int*, int, int, int);
  int hinv(int*, int, int, int, int);
}

template<class T> FitsHcompressm<T>::FitsHcompressm(FitsFile* fits) 
  : FitsCompressm<T>(fits)
{
  // hcompress parameters
  scale_ = 1;
  smooth_ = 0;
  char name[] = "ZNAME ";
  char val[] = "ZVAL ";
  for (int ii=0; ii<9; ii++) {
    name[5] = '0'+ii;
    val[4] = '0'+ii;
    if (fits->find(name)) {
      char* which = fits->getString(name);
      if (!strncmp(which,"SCALE",4))
	scale_ = fits->getInteger(val,32);
      else if (!strncmp(which,"SMOOTH",4))
	smooth_ = fits->getInteger(val,4);
    }
  }

  // create header
  if (!FitsCompressm<T>::initHeader(fits))
    return;

  // uncompress
  if (!inflate(fits))
    return;

  // swap if needed
  FitsCompressm<T>::swapBytes();

  // all done
  FitsCompressm<T>::valid_ = 1;
}

template <class T> int FitsHcompressm<T>::inflate(FitsFile* fits)
{
  FitsHead* srcHead = fits->head();
  FitsTableHDU* srcHDU = (FitsTableHDU*)(srcHead->hdu());

  T* dest = new T[FitsCompressm<T>::size_];
  if (!dest) {
    if (DebugCompress)
      cerr << "HCOMPRESS Internal Error: unable to allocate memory" << endl;
    return 0;
  }

  // init image
  memset(dest, 0, FitsCompressm<T>::size_*sizeof(T));

  // src
  char* sdata = (char*)fits->data(); 
  char* sptr = sdata;
  int heap = srcHDU->realbytes();
  int rowlen = srcHDU->width();
  int rows = srcHDU->rows();

  // dest
  T* dptr = dest;
  int ocnt = FitsCompressm<T>::ww_*FitsCompressm<T>::hh_*FitsCompressm<T>::dd_;

  for (int ii=0; ii<rows; ii++, sptr+=rowlen) {
    // for memory models that support internal paging
    // all we need is just one row
    sptr = fits->page(sptr, rowlen);

    double zs = FitsCompressm<T>::bscale_;
    if (FitsCompressm<T>::zscale_)
      zs = FitsCompressm<T>::zscale_->value(sptr,0);

    double zz = FitsCompressm<T>::bzero_;
    if (FitsCompressm<T>::zzero_)
      zz = FitsCompressm<T>::zzero_->value(sptr,0);

    if (FitsCompressm<T>::compress_) {
      int icnt=0;
      unsigned char* ibuf = 
	(unsigned char*)((FitsBinColumnArray*)FitsCompressm<T>::compress_)->get(sdata+heap, sptr, &icnt);
      int obuf[ocnt];

      unsigned char* ptr = ibuf;
      // magic numbers
      if (*ptr++ != 0xDD || *ptr++ != 0x99) {
	if (DebugCompress)
	  cerr << "HCOMPRESS Internal Error: bad magic numbers" << endl;
	return 0;
      }

      // size,scale,checksum,nbitplanes
      int nx = readint(ptr);
      ptr += sizeof(int);
      int ny = readint(ptr);
      ptr += sizeof(int); // ny
      int scale = readint(ptr);
      ptr += sizeof(int); // scale
      int sum = readint(ptr);
      ptr += sizeof(int); // checksum
      int sum2 = readint(ptr);
      ptr += sizeof(int); // checksum

      // nbitplanes
      unsigned char nbp[3];
      nbp[0] = *ptr++;
      nbp[1] = *ptr++;
      nbp[2] = *ptr++;

      // decode
      if (!dodecode(&ptr, obuf, nx, ny, nbp)) {
	if (DebugCompress)
	  cerr << "HCOMPRESS Internal Error: bad inflate result" << endl;
	return 0;
      }
      obuf[0] = sum;

      // undigitize
      undigitize(obuf, nx, ny, scale);

      // hinv
      if (!hinv(obuf, nx, ny, 0, scale)) {
	if (DebugCompress)
	  cerr << "HCOMPRESS Internal Error: bad transform" << endl;
	return 0;
      }

      // scale
      if (zs != 1 || zz != 0)
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk]*zs + zz;
      else
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk];
    }

    if (FitsCompressm<T>::uncompress_) {
      int ocnt=0;
      T* obuf = 
	(T*)(((FitsBinColumnArray*)FitsCompressm<T>::uncompress_)->get(sdata+heap, sptr, &ocnt));

      if (zs != 1 || zz != 0)
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk]*zs + zz;
      else
	for (int kk=0; kk<ocnt; kk++, dptr++)
	  *dptr = obuf[kk];
    }
  }

  // for memory models that support internal paging
  fits->resetpage();

  FitsCompressm<T>::data_ = dest;
  return 1;
}

template <class T> int FitsHcompressm<T>::readint(unsigned char* ptr)
{
  if (FitsCompressm<T>::byteswap_) {
    int rr;
    swap4((char*)ptr,(char*)(&rr));
    return rr;
  }
  else
    return *((int*)ptr);
}

template class FitsHcompressm<unsigned char>;
template class FitsHcompressm<short>;
template class FitsHcompressm<unsigned short>;
template class FitsHcompressm<int>;
template class FitsHcompressm<long long>;
template class FitsHcompressm<float>;
template class FitsHcompressm<double>;
