// Copyright (C) 1999-2007
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;

#include "rice.h"

extern "C" {
  int fits_rdecomp (unsigned char *c,
		    int clen,
		    unsigned int array[],
		    int nx,
		    int nblock);
  void ffpmsg(const char* str) 
  {
    if (DebugCompress)
      cerr << str << endl;
  }
}

template<class T> FitsRicem<T>::FitsRicem(FitsFile* fits) 
  : FitsCompressm<T>(fits)
{
  // rice parameters
  block_ = 32;
  noisebit_ = 4;
  char name[] = "ZNAME ";
  char val[] = "ZVAL ";
  for (int ii=0; ii<9; ii++) {
    name[5] = '0'+ii;
    val[4] = '0'+ii;
    if (fits->find(name)) {
      char* which = fits->getString(name);
      if (!strncmp(which,"BLOCK",4))
	block_ = fits->getInteger(val,32);
      else if (!strncmp(which,"NOISEBIT",4))
	noisebit_ = fits->getInteger(val,4);
    }
  }

  // create header
  if (!FitsCompressm<T>::initHeader(fits))
    return;

  // uncompress
  if (!inflate(fits))
    return;

  // swap if needed
  FitsCompressm<T>::swapBytes();

  // all done
  FitsCompressm<T>::valid_ = 1;
}

template <class T> int FitsRicem<T>::inflate(FitsFile* fits)
{
  FitsHead* srcHead = fits->head();
  FitsTableHDU* srcHDU = (FitsTableHDU*)(srcHead->hdu());

  T* dest = new T[FitsCompressm<T>::size_];
  if (!dest) {
    if (DebugCompress)
      cerr << "RICE Internal Error: unable to allocate memory" << endl;
    return 0;
  }

  // init image
  memset(dest, 0, FitsCompressm<T>::size_*sizeof(T));

  // src
  char* sdata = (char*)fits->data(); 
  char* sptr = sdata;
  int heap = srcHDU->realbytes();
  int rowlen = srcHDU->width();
  int rows = srcHDU->rows();

  // dest
  int iistart =0;
  int iistop =FitsCompressm<T>::ww_;
  if (iistop > FitsCompressm<T>::width_)
    iistop = FitsCompressm<T>::width_;

  int jjstart =0;
  int jjstop =FitsCompressm<T>::hh_;
  if (jjstop > FitsCompressm<T>::height_)
    jjstop = FitsCompressm<T>::height_;

  int kkstart =0;
  int kkstop =FitsCompressm<T>::dd_;

  for (int aa=0; aa<rows; aa++, sptr+=rowlen) {
    // for memory models that support internal paging
    // all we need is just one row
    sptr = fits->page(sptr, rowlen);

    double zs = FitsCompressm<T>::bscale_;
    if (FitsCompressm<T>::zscale_)
      zs = FitsCompressm<T>::zscale_->value(sptr,0);

    double zz = FitsCompressm<T>::bzero_;
    if (FitsCompressm<T>::zzero_)
      zz = FitsCompressm<T>::zzero_->value(sptr,0);

    if (FitsCompressm<T>::compress_) {
      int icnt=0;
      unsigned char* ibuf = 
	(unsigned char*)((FitsBinColumnArray*)FitsCompressm<T>::compress_)->get(sdata+heap, sptr, &icnt);

      int ocnt = FitsCompressm<T>::ww_ * FitsCompressm<T>::hh_ * 
	FitsCompressm<T>::dd_;
      unsigned int obuf[ocnt];

      if (fits_rdecomp(ibuf, icnt, obuf, ocnt, block_)) {
	if (DebugCompress)
	  cerr << "RICE Internal Error: bad inflate result" << endl;
	return 0;
      }

      int ll=0;
      if (zs != 1 || zz != 0)
	for (int kk=kkstart; kk<kkstop; kk++)
	  for (int jj=jjstart; jj<jjstop; jj++)
	    for (int ii=iistart; ii<iistop; ii++,ll++)
	      dest[kk*FitsCompressm<T>::width_*FitsCompressm<T>::height_ +
		   jj*FitsCompressm<T>::width_ +
		   ii] = obuf[ll]*zs + zz;
      else
	for (int kk=kkstart; kk<kkstop; kk++)
	  for (int jj=jjstart; jj<jjstop; jj++)
	    for (int ii=iistart; ii<iistop; ii++,ll++)
	      dest[kk*FitsCompressm<T>::width_*FitsCompressm<T>::height_ + 
		   jj*FitsCompressm<T>::width_ +
		   ii] = obuf[ll];
    }

    if (FitsCompressm<T>::uncompress_) {
      int ocnt=0;
      T* obuf = 
	(T*)(((FitsBinColumnArray*)FitsCompressm<T>::uncompress_)->get(sdata+heap, sptr, &ocnt));

      int ll=0;
      if (zs != 1 || zz != 0)
	for (int kk=kkstart; kk<kkstop; kk++)
	  for (int jj=jjstart; jj<jjstop; jj++)
	    for (int ii=iistart; ii<iistop; ii++,ll++)
	      dest[kk*FitsCompressm<T>::width_*FitsCompressm<T>::height_ +
		   jj*FitsCompressm<T>::width_ +
		   ii] = obuf[ll]*zs + zz;
      else
	for (int kk=kkstart; kk<kkstop; kk++)
	  for (int jj=jjstart; jj<jjstop; jj++)
	    for (int ii=iistart; ii<iistop; ii++,ll++)
	      dest[kk*FitsCompressm<T>::width_*FitsCompressm<T>::height_ + 
		   jj*FitsCompressm<T>::width_ +
		   ii] = obuf[ll];
    }

    // tiles may not be an even multiple of the image size
    iistart += FitsCompressm<T>::ww_;
    iistop += FitsCompressm<T>::ww_;
    if (iistop > FitsCompressm<T>::width_)
      iistop = FitsCompressm<T>::width_;

    if (iistart >= FitsCompressm<T>::width_) {
      iistart = 0;
      iistop = FitsCompressm<T>::ww_;
      if (iistop > FitsCompressm<T>::width_)
	iistop = FitsCompressm<T>::width_;

      jjstart += FitsCompressm<T>::hh_;
      jjstop += FitsCompressm<T>::hh_;
      if (jjstop > FitsCompressm<T>::height_)
	jjstop = FitsCompressm<T>::height_;

      if (jjstart >= FitsCompressm<T>::height_) {
	jjstart = 0;
	jjstop = FitsCompressm<T>::hh_;
	if (jjstop > FitsCompressm<T>::height_)
	  jjstop = FitsCompressm<T>::height_;

	kkstart += FitsCompressm<T>::dd_;
	kkstop += FitsCompressm<T>::dd_;
      }
    }
  }

  // for memory models that support internal paging
  fits->resetpage();

  FitsCompressm<T>::data_ = dest;
  return 1;
}

template class FitsRicem<unsigned char>;
template class FitsRicem<short>;
template class FitsRicem<unsigned short>;
template class FitsRicem<int>;
template class FitsRicem<long long>;
template class FitsRicem<float>;
template class FitsRicem<double>;
